
###################################
##  Ian's common.mk             ###
##  (based on humper.mk)        ###
###################################

ifndef ROOTDIR
ROOTDIR := .
endif

# Where the base directories should be
BIN        := bin
BUILT      := built
INC        := include
DEP        := depends
BINDIR     := $(ROOTDIR)/$(BIN)
OBJDIR 	   := $(BUILT)
INCLUDEDIR := $(ROOTDIR)/$(INC)
DEPDIR     := $(DEP)
FASTDEPS   := $(ROOTDIR)/scripts/fastdep.pl

INCLUDEFILES := $(ROOTDIR)/$(INC)/*.h
MAKEFILEDEPS := $(ROOTDIR)/common.mk Makefile

#Don't delete object files
.SECONDARY:


# Include the OS settings
#ifndef OS
OS := $(shell uname)
#endif

include $(ROOTDIR)/config/$(OS).mk

CFLAGS    += $(C_INCLUDE_FLAG)$(INCLUDEDIR)
LDFLAGS   += $(LD_LIBDIR_FLAG)$(ROOTDIR)/$(BIN)

TEMP      := $(addprefix $(LD_LIBLINK_PREFIX), $(LIBRARIES))
LDFLAGS   += $(addsuffix $(LD_LIBLINK_SUFFIX), $(TEMP))

COMMON_OBJS := $(COMMON_FILES)
COMMON_OBJS := $(addprefix $(OBJDIR)/, $(COMMON_OBJS))
COMMON_OBJS := $(addsuffix $(OBJSUFFIX), $(COMMON_OBJS))

DEPS    := $(COMMON_FILES) $(TARGETS)
DEPS    := $(addprefix $(DEPDIR)/, $(DEPS))
DEPS    := $(addsuffix .depend, $(DEPS))

BINARY      := $(addprefix $(BINDIR)/, $(TARGETS))
BINARY      := $(addsuffix $(BINSUFFIX), $(BINARY))

ifdef RELEASE
CFLAGS   += $(C_RELEASE_FLAG)
LDFLAGS  += $(LD_RELEASE_FLAG)
else
ifdef PROFILE
CFLAGS   += $(C_PROFILE_FLAG)
LDFLAGS  += $(LD_PROFILE_FLAG)
else
CFLAGS   += $(C_DEBUG_FLAG)
LDFLAGS  += $(LD_DEBUG_FLAG)
endif
endif

ifdef GLUT
LIBRARIES += GLUTLIB
endif

all: arch

debug:
	@$(MAKE) --no-print-directory

profile:
	@$(MAKE) PROFILE=1 --no-print-directory

release:
	@$(MAKE) RELEASE=1 --no-print-directory

arch: $(PRECOMP) makedirs dep

recurse: $(BINARY)

dep: $(DEPS)
	@$(MAKE) --no-print-directory recurse INCLUDEDEPS=1

makedirs:
	@if test ! -d $(OBJDIR); then $(MKDIR) $(OBJDIR); fi
	@if test ! -d $(BINDIR); then $(MKDIR) $(BINDIR); fi
	@if test ! -d $(DEPDIR); then $(MKDIR) $(DEPDIR); fi

.SUFFIXES : $OBJSUFFIX .c .cpp

ifdef INCLUDEDEPS
include $(DEPS)
else
$(DEPDIR)/%.depend: %.c
	@$(MAKE_DEPDIR)
	@$(ECHO) "Rebuilding dependencies for $<"
	@$(PERL) $(FASTDEPS) -I. -I$(INCLUDEDIR) --obj-prefix='$(OBJDIR)/' $< > $@

$(DEPDIR)/%.depend: %.cpp
	@$(MAKE_DEPDIR)
	@$(ECHO) "Rebuilding dependencies for $<"
	@$(PERL) $(FASTDEPS) -I. -I$(INCLUDEDIR) --obj-prefix='$(OBJDIR)/' $< > $@
endif

$(OBJDIR)/%$(OBJSUFFIX): %.c
	@$(CC) $(CFLAGS) $(C_OUTPUT_FLAG)$@ $(C_COMPILE_FLAG) $<

$(OBJDIR)/%$(OBJSUFFIX): %.cpp
	@$(CC) $(CFLAGS) $(C_OUTPUT_FLAG)$@ $(C_COMPILE_FLAG) $<

$(BINDIR)/%$(BINSUFFIX): $(OBJDIR)/%$(OBJSUFFIX) $(COMMON_OBJS)
	@$(ECHO) Building $@
	@$(LD) $(LDFLAGS) $(LD_OUTPUT_FLAG)$@ $< $(COMMON_OBJS) $(LD_LIBRARIES_LINK)


CLEANFILES     := $(addsuffix .pdb, $(TARGETS)) \
	  	  $(addsuffix .map, $(TARGETS)) \
	  	  $(addsuffix .exp, $(TARGETS)) \
	  	  $(addsuffix .pbi, $(TARGETS)) \
	  	  $(addsuffix .pbo, $(TARGETS)) \
	  	  $(addsuffix .pbt, $(TARGETS)) \
	  	  $(addsuffix ._xe, $(TARGETS)) \
	  	  $(addsuffix .lib, $(TARGETS))
CLEANFILES     := $(addprefix $(BINDIR)/, $(CLEANFILES))

clean:
	@rm -rf $(OBJDIR) $(BINARY) $(SLOP) *~  \
		$(CLEANFILES) $(DEPDIR) vc60.pdb 

clobber:
	@rm -rf $(OBJDIR) $(BINDIR) $(SLOP) *~  \
		$(DEPDIR) vc60.pdb 


