/*
        Handles setup of gl window context and pbuffer creation.

        2/10/04  Made changes so decision to go down ATI/NV code path is
        determined at runtime. Formerly done at compile time. (Kayvon)
*/


#include <windows.h>
#include <assert.h>
#include <GL/gl.h>
#include <stdio.h>

#include "gpubench.h"


DECLARE_HANDLE(HPBUFFERARB);

HWND  hwnd;
HDC   hdc;
HGLRC hglrc;

HPBUFFERARB hpbuffer;
HDC         hpbufferdc;
HGLRC       hpbufferglrc;

#define WINDOW_NAME "GPUBench"

typedef HPBUFFERARB (APIENTRY * PFNWGLCREATEPBUFFERARB)  (HDC hDC,
                                      int iPixelFormat,
                                      int iWidth,
                                      int iHeight,
                                      const int *piAttribList);
typedef HDC (APIENTRY * PFNWGLGETPBUFFERDCARB) (HPBUFFERARB hPbuffer);
typedef BOOL (APIENTRY * PFNWGLCHOOSEPIXELFORMATARB) (HDC hdc,
                                                      const GLint *piAttribIList,
                                                      const GLfloat *pfAttribFList,
                                                      GLuint nMaxFormats,
                                                      GLint *piFormats,
                                                      GLuint *nNumFormats);
typedef BOOL (APIENTRY * PFNWGLBINDTEXIMAGEARB) (HPBUFFERARB hPbuffer, int iBuffer);
typedef BOOL (APIENTRY * PFNWGLRELEASETEXIMAGEARB) (HPBUFFERARB hPbuffer, int iBuffer);
typedef int  (APIENTRY * PFNWGLRELEASEPBUFFERDCARB) (HPBUFFERARB hPbuffer, HDC hdc);
typedef BOOL (APIENTRY * PFNWGLDESTROYPBUFFERARB) (HPBUFFERARB hPbuffer);

PFNWGLCHOOSEPIXELFORMATARB wglChoosePixelFormatARB;
PFNWGLCREATEPBUFFERARB wglCreatePbufferARB;
PFNWGLGETPBUFFERDCARB  wglGetPbufferDCARB;
PFNWGLBINDTEXIMAGEARB  wglBindTexImageARB;
PFNWGLRELEASETEXIMAGEARB  wglReleaseTexImageARB;
PFNWGLRELEASEPBUFFERDCARB wglReleasePbufferDCARB;
PFNWGLDESTROYPBUFFERARB wglDestroyPbufferARB;

void (*draw) (void) = NULL;

#define WGL_FULL_ACCELERATION_ARB               0x2027
#define WGL_COLOR_BITS_ARB                      0x2014
#define WGL_COLOR_BITS_ARB                      0x2014
#define WGL_DEPTH_BITS_ARB                      0x2022
#define WGL_STENCIL_BITS_ARB                    0x2023
#define WGL_DOUBLE_BUFFER_ARB                   0x2011
#define WGL_ACCELERATION_ARB                    0x2003
#define WGL_ALPHA_BITS_ARB                      0x201B


#define WGL_NUMBER_PIXEL_FORMATS_ARB        0x2000
#define WGL_DRAW_TO_WINDOW_ARB              0x2001
#define WGL_DRAW_TO_BITMAP_ARB              0x2002
#define WGL_ACCELERATION_ARB                0x2003
#define WGL_NEED_PALETTE_ARB                0x2004
#define WGL_NEED_SYSTEM_PALETTE_ARB         0x2005
#define WGL_SWAP_LAYER_BUFFERS_ARB          0x2006
#define WGL_SWAP_METHOD_ARB                 0x2007
#define WGL_NUMBER_OVERLAYS_ARB             0x2008
#define WGL_NUMBER_UNDERLAYS_ARB            0x2009
#define WGL_TRANSPARENT_ARB                 0x200A
#define WGL_TRANSPARENT_RED_VALUE_ARB       0x2037
#define WGL_TRANSPARENT_GREEN_VALUE_ARB     0x2038
#define WGL_TRANSPARENT_BLUE_VALUE_ARB      0x2039
#define WGL_TRANSPARENT_ALPHA_VALUE_ARB     0x203A
#define WGL_TRANSPARENT_INDEX_VALUE_ARB     0x203B
#define WGL_SHARE_DEPTH_ARB                 0x200C
#define WGL_SHARE_STENCIL_ARB               0x200D
#define WGL_SHARE_ACCUM_ARB                 0x200E
#define WGL_SUPPORT_GDI_ARB                 0x200F
#define WGL_SUPPORT_OPENGL_ARB              0x2010
#define WGL_DOUBLE_BUFFER_ARB               0x2011
#define WGL_STEREO_ARB                      0x2012
#define WGL_PIXEL_TYPE_ARB                  0x2013
#define WGL_COLOR_BITS_ARB                  0x2014
#define WGL_RED_BITS_ARB                    0x2015
#define WGL_RED_SHIFT_ARB                   0x2016
#define WGL_GREEN_BITS_ARB                  0x2017
#define WGL_GREEN_SHIFT_ARB                 0x2018
#define WGL_BLUE_BITS_ARB                   0x2019
#define WGL_BLUE_SHIFT_ARB                  0x201A
#define WGL_ALPHA_BITS_ARB                  0x201B
#define WGL_ALPHA_SHIFT_ARB                 0x201C
#define WGL_ACCUM_BITS_ARB                  0x201D
#define WGL_ACCUM_RED_BITS_ARB              0x201E
#define WGL_ACCUM_GREEN_BITS_ARB            0x201F
#define WGL_ACCUM_BLUE_BITS_ARB             0x2020
#define WGL_ACCUM_ALPHA_BITS_ARB            0x2021
#define WGL_DEPTH_BITS_ARB                  0x2022
#define WGL_STENCIL_BITS_ARB                0x2023
#define WGL_AUX_BUFFERS_ARB                 0x2024
#define WGL_NO_ACCELERATION_ARB             0x2025
#define WGL_GENERIC_ACCELERATION_ARB        0x2026
#define WGL_FULL_ACCELERATION_ARB           0x2027
#define WGL_SWAP_EXCHANGE_ARB               0x2028
#define WGL_SWAP_COPY_ARB                   0x2029
#define WGL_SWAP_UNDEFINED_ARB              0x202A
#define WGL_TYPE_RGBA_ARB                   0x202B
#define WGL_TYPE_COLORINDEX_ARB             0x202C
#define WGL_BIND_TO_TEXTURE_RGBA_ARB        0x2071
#define WGL_TYPE_RGBA_FLOAT_ATI             0x21A0

#define WGL_DRAW_TO_PBUFFER_ARB                             0x202D
#define WGL_FLOAT_COMPONENTS_NV        0x20B0
#define WGL_BIND_TO_TEXTURE_RECTANGLE_FLOAT_R_NV 0x20B1
#define WGL_BIND_TO_TEXTURE_RECTANGLE_FLOAT_RG_NV 0x20B2
#define WGL_BIND_TO_TEXTURE_RECTANGLE_FLOAT_RGB_NV 0x20B3
#define WGL_BIND_TO_TEXTURE_RECTANGLE_FLOAT_RGBA_NV 0x20B4
#define WGL_TEXTURE_FLOAT_R_NV         0x20B5
#define WGL_TEXTURE_FLOAT_RG_NV        0x20B6
#define WGL_TEXTURE_FLOAT_RGB_NV       0x20B7
#define WGL_TEXTURE_FLOAT_RGBA_NV      0x20B8

#define WGL_TEXTURE_FORMAT_ARB                              0x2072
#define WGL_TEXTURE_TARGET_ARB                              0x2073
#define WGL_TEXTURE_RGBA_ARB                                0x2076
#define WGL_TEXTURE_2D_ARB                                  0x207A

#define WGL_TEXTURE_RECTANGLE_NV                            0x20A2

#define WGL_FRONT_LEFT_ARB                                  0x2083
#define WGL_FRONT_ARB                       WGL_FRONT_LEFT_ARB

#define WGL_BACK_LEFT_ARB                                   0x2085
#define WGL_BACK_ARB                        WGL_BACK_LEFT_ARB

static LONG WINAPI
MainWndProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{

  switch ( uMsg ) {
  case WM_PAINT:
    if (draw)
      draw();
    break;
  case WM_CLOSE:
    exit( 0 );
    break;
  }

  return DefWindowProc( hWnd, uMsg, wParam, lParam );
}

static BOOL
bSetupPixelFormat( HDC hdc )
{
    PIXELFORMATDESCRIPTOR *ppfd;
    PIXELFORMATDESCRIPTOR pfd = {
      sizeof(PIXELFORMATDESCRIPTOR),  //  size of this pfd
      1,                              // version number
      PFD_DRAW_TO_WINDOW |            // support window
      PFD_SUPPORT_OPENGL |            // support OpenGL
      PFD_DOUBLEBUFFER,               // double buffered
      PFD_TYPE_RGBA,                  // RGBA type
      32,                             // 32-bit color depth
      0, 0, 0, 0, 0, 0,               // color bits ignored
      8,                              // standard 8bit alpha buffer
      0,                              // alpha shift bit ignored
      0,                              // no accumulation buffer
      0, 0, 0, 0,                     // accum bits ignored
      24,                             // set depth buffer
      0,                              // no stencil buffer
      0,                              // no auxiliary buffer
      PFD_MAIN_PLANE,                 // main layer
      0,                              // reserved
      0, 0, 0                         // layer masks ignored
    };
    int pixelformat;

    ppfd = &pfd;
    pixelformat = ChoosePixelFormat( hdc, ppfd );

    if ( pixelformat == 0 )
      {
        fprintf( stderr, "ChoosePixelFormat failed" );
        return FALSE;
      }
    if ( !SetPixelFormat( hdc, pixelformat, ppfd ) )
      {
        fprintf( stderr, "SetPixelFormat failed" );
        return FALSE;
      }

    return TRUE;
}

/* standard front/back flipping */
void gpubench_swapBuffers(void) {
  SwapBuffers(hdc);
}


void  gpubench_createWindow(int w, int h, void (*draw_func) (void),
                    char name[]) {

  HINSTANCE hinstance;
  WNDCLASS wc;
  DWORD window_style;
  char wndname[2048];
  MSG msg;

  hinstance = GetModuleHandle( NULL );

  if ( !GetClassInfo(hinstance, WINDOW_NAME, &wc) )
    {
      wc.style = CS_OWNDC;
      wc.lpfnWndProc = (WNDPROC) MainWndProc;
      wc.cbClsExtra = 0;
      wc.cbWndExtra = 0;
      wc.hInstance = hinstance;
      wc.hIcon = LoadIcon( NULL, IDI_APPLICATION );
      wc.hCursor = LoadCursor( NULL, IDC_ARROW );
      wc.hbrBackground = NULL;
      wc.lpszMenuName = NULL;
      wc.lpszClassName = WINDOW_NAME;

      assert (RegisterClass( &wc ));
    }

  window_style = ( WS_CLIPSIBLINGS | WS_CLIPCHILDREN );
  window_style |= WS_POPUP;

  assert (name);
  assert (strlen(WINDOW_NAME) + strlen(name) + 3 < 2048);
  sprintf (wndname, "%s: %s", WINDOW_NAME, name);

  hwnd = CreateWindow( WINDOW_NAME, wndname,
                       window_style,
                       0,
                       0,
                       w, h,
                       NULL, NULL, hinstance, NULL );

  assert (hwnd);

  // ShowWindow( hwnd, SW_SHOWNORMAL );
  hdc = GetDC( hwnd);

  assert(bSetupPixelFormat( hdc ) );

  hglrc = wglCreateContext( hdc );
  wglMakeCurrent( hdc, hglrc );

  draw = draw_func;

  /* Clear message queue */
  while( PeekMessage( &msg, NULL, 0, 0, PM_REMOVE )) {
    TranslateMessage(&msg);
    DispatchMessage(&msg);
  }
}

void gpubench_getprogramiv(GLenum p, GLenum e, int *i) {
   PFNGLGETPROGRAMIVARBPROC glGetProgramiv;
   glGetProgramiv = (PFNGLGETPROGRAMIVARBPROC)
      wglGetProcAddress("glGetProgramivARB");
   assert(glGetProgramiv);
   glGetProgramiv(p, e, i);
   return;
}

void gpubench_showWindow(void) {
  ShowWindow(hwnd, SW_SHOWNORMAL);
}

void gpubench_hideWindow(void) {
  ShowWindow(hwnd, SW_HIDE);
}

void gpubench_makeCurrentWindow(void) {
  assert(wglMakeCurrent( hdc, hglrc ));
}

void gpubench_makeCurrentPbuffer(void) {
  assert(wglMakeCurrent( hpbufferdc, hpbufferglrc ));
}

/* bind pbuffer to a texture unit */
void gpubench_bindPbuffer(int front_or_back) {
  assert(wglBindTexImageARB(hpbuffer, front_or_back?WGL_FRONT_ARB:WGL_BACK_ARB));
}

void gpubench_bindPbufferAux(int which) {
  assert(wglBindTexImageARB(hpbuffer, which));
}
void gpubench_releasePbufferAux(int which) {
  assert(wglReleaseTexImageARB(hpbuffer, which));
}

/* release bound pbuffer from texture */
void gpubench_releasePbuffer(int front_or_back) {
  assert(wglReleaseTexImageARB(hpbuffer, front_or_back?WGL_FRONT_ARB:WGL_BACK_ARB));
}

void gpubench_destroyPbuffer(void) {
  assert(wglMakeCurrent( hpbufferdc, NULL ));
  assert(wglDeleteContext( hpbufferglrc ));
  assert(wglReleasePbufferDCARB(hpbuffer, hpbufferdc));
  assert(wglDestroyPbufferARB(hpbuffer));
}

/*
 * takes render target (front face if 'draw_to_front' == TRUE) and binds to
 * active texture unit, then flips the render target surface to the
 * front/backbuffer
 */
int gpubench_bindAndSwap(int draw_to_front) {

  // KAYVON: should this be here?
  glFinish();

  if (draw_to_front) {
    //printf("Moving front buffer to TEX and drawing to BACK!\n");
    gpubench_releasePbuffer(GL_FALSE);
    gpubench_bindPbuffer(GL_TRUE);
    glDrawBuffer(GL_BACK);
    glReadBuffer(GL_BACK);
    return 0;
  } else {
    //printf("Moving back buffer to TEX and drawing to FRONT!\n");
    gpubench_releasePbuffer(GL_TRUE);
    gpubench_bindPbuffer(GL_FALSE);
    glDrawBuffer(GL_FRONT);
    glReadBuffer(GL_FRONT);
    return 1;
  }
}



int gpubench_createFloatPbuffer(int w, int h, int doublebuffer) {
  return gpubench_createFloatPbufferAuxComponents(w, h, doublebuffer, 0, 4);
}

int gpubench_createFloatPbufferComponents(int w, int h, int doublebuffer,
                                          int components) {
  return gpubench_createFloatPbufferAuxComponents(w, h,
                                                  doublebuffer,
                                                  0, components);
}


int gpubench_createFloatPbufferAux(int w, int h, int doublebuffer, int auxBuffers) {
  return gpubench_createFloatPbufferAuxComponents(w, h,
                                                  doublebuffer,
                                                  auxBuffers, 4);
}

int gpubench_createFloatPbufferAuxComponents(int w, int h, int doublebuffer,
                                             int auxBuffers, int components) {
   return gpubench_createFloatPbufferAll(w, h,
                                         doublebuffer,
                                         auxBuffers, components, 0);
}


int gpubench_createFloatPbufferAll(int w, int h, int doublebuffer,
                                   int auxBuffers, int components, int depth) {

#define ADD_PAIR(i, array, val1, val2) \
  { unsigned int index = (i);  \
    array[2*(index)] = (val1); \
    array[2*(index)+1] = (val2); }

  static bool loadedSyms;
  int pbufferFlags[30];
  int iAttributes[30];
  float fAttributes[] = {0, 0};
  unsigned int numFormats, i = 0;
  BOOL status;
  int pixelformat;

  assert( !(doublebuffer && auxBuffers) );
  assert( components > 0 && components < 5);

  if (!loadedSyms) {
    loadedSyms = true;
    GETPROCADDR(PFNWGLBINDTEXIMAGEARB, wglBindTexImageARB);
    GETPROCADDR(PFNWGLCHOOSEPIXELFORMATARB, wglChoosePixelFormatARB);
    GETPROCADDR(PFNWGLCREATEPBUFFERARB, wglCreatePbufferARB);
    GETPROCADDR(PFNWGLDESTROYPBUFFERARB, wglDestroyPbufferARB);
    GETPROCADDR(PFNWGLGETPBUFFERDCARB, wglGetPbufferDCARB);
    GETPROCADDR(PFNWGLRELEASEPBUFFERDCARB, wglReleasePbufferDCARB);
    GETPROCADDR(PFNWGLRELEASETEXIMAGEARB, wglReleaseTexImageARB);

    assert (wglCreatePbufferARB && wglGetPbufferDCARB &&
            wglChoosePixelFormatARB && wglBindTexImageARB &&
            wglReleaseTexImageARB && wglReleasePbufferDCARB &&
            wglDestroyPbufferARB);
  }

  if (gpubench_isNV()) {

    static int format[5] = {0,
                            WGL_TEXTURE_FLOAT_R_NV,
                            WGL_TEXTURE_FLOAT_RG_NV,
                            WGL_TEXTURE_FLOAT_RGB_NV,
                            WGL_TEXTURE_FLOAT_RGBA_NV};
    static int bind[5] = {0,
                          WGL_BIND_TO_TEXTURE_RECTANGLE_FLOAT_R_NV,
                          WGL_BIND_TO_TEXTURE_RECTANGLE_FLOAT_RG_NV,
                          WGL_BIND_TO_TEXTURE_RECTANGLE_FLOAT_RGB_NV,
                          WGL_BIND_TO_TEXTURE_RECTANGLE_FLOAT_RGBA_NV};

    ADD_PAIR(0,pbufferFlags,WGL_TEXTURE_FORMAT_ARB, format[components]);
    ADD_PAIR(1,pbufferFlags,WGL_TEXTURE_TARGET_ARB, WGL_TEXTURE_RECTANGLE_NV);
    ADD_PAIR(2,pbufferFlags, 0, 0);

    ADD_PAIR(i++, iAttributes, WGL_FLOAT_COMPONENTS_NV, GL_TRUE);
    ADD_PAIR(i++, iAttributes, bind[components],        GL_TRUE);
  }  else {

    ADD_PAIR(0, pbufferFlags, WGL_TEXTURE_FORMAT_ARB, WGL_TEXTURE_RGBA_ARB);
    ADD_PAIR(1, pbufferFlags, WGL_TEXTURE_TARGET_ARB, WGL_TEXTURE_2D_ARB);
    ADD_PAIR(2, pbufferFlags, 0, 0);

    ADD_PAIR(i++, iAttributes, WGL_BIND_TO_TEXTURE_RGBA_ARB, TRUE);
    ADD_PAIR(i++, iAttributes, WGL_PIXEL_TYPE_ARB, WGL_TYPE_RGBA_FLOAT_ATI);
  }

  ADD_PAIR(i++, iAttributes,  WGL_SUPPORT_OPENGL_ARB,  TRUE);
  ADD_PAIR(i++, iAttributes,  WGL_DRAW_TO_PBUFFER_ARB, TRUE);
  ADD_PAIR(i++, iAttributes,  WGL_ACCELERATION_ARB,    WGL_FULL_ACCELERATION_ARB);
#if 0
  /*
   * For some reason, 3Dlabs cards choke if this is set to a value higher
   * than 64.
   */
  ADD_PAIR(i++, iAttributes,  WGL_COLOR_BITS_ARB,      components*32);
#endif

  /*
   * Note: 3Dlabs cards are very picky about this.  If you try to specify a
   * less than 4 component format by having the channel present, but with a
   * value of zero, the driver will refuse.
   */
  ADD_PAIR(i++, iAttributes,  WGL_RED_BITS_ARB,        32);
  if (components > 1) {
     ADD_PAIR(i++, iAttributes,  WGL_GREEN_BITS_ARB,   32);
     if (components > 2) {
        ADD_PAIR(i++, iAttributes,  WGL_BLUE_BITS_ARB, 32);
        if (components > 3) {
           ADD_PAIR(i++, iAttributes,  WGL_ALPHA_BITS_ARB, 32);
        }
     }
  }

  /*
   * Note: 3DLabs cards reject requests that mention depth, but ask for zero
   * bits.
   */

  if (depth > 0) {
     ADD_PAIR(i++, iAttributes, WGL_DEPTH_BITS_ARB,       depth);
  }
  ADD_PAIR(i++, iAttributes, WGL_AUX_BUFFERS_ARB,      auxBuffers);
  ADD_PAIR(i++, iAttributes, WGL_DOUBLE_BUFFER_ARB,    doublebuffer);
  ADD_PAIR(i++, iAttributes, 0, 0);

  status = wglChoosePixelFormatARB(hdc, iAttributes, fAttributes, 1,
                                   &pixelformat, &numFormats);
  if (!status) {
     static char str[2048];
     int err = GetLastError();

     if (FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM|FORMAT_MESSAGE_IGNORE_INSERTS,
                       NULL, err, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                       str, sizeof str, NULL) == 0) {
        sprintf(str, "Unknown error %d (0x%x)\n", err, err);
     }
     fprintf(stderr, "wglChoosePixelFormatARB failed: %s\n", str);
     return FALSE;
  }
  if (numFormats == 0) {
     fprintf(stderr, "No float%d pbuffer formats available.\n", components);
     return FALSE;
  }

  hpbuffer = wglCreatePbufferARB(hdc,
                                 pixelformat,
                                 w, h,
                                 pbufferFlags);
  if (hpbuffer == NULL) {
     static char str[2048];
     int err = GetLastError();

     if (FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM|FORMAT_MESSAGE_IGNORE_INSERTS,
                       NULL, err, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                       str, sizeof str, NULL) == 0) {
        sprintf(str, "Unknown error %d (0x%x)\n", err, err);
     }
     fprintf(stderr, "wglCreatePbufferARB failed: %s\n", str);
     return FALSE;
  }

  hpbufferdc = wglGetPbufferDCARB (hpbuffer);
  assert (hpbufferdc);

  hpbufferglrc = wglCreateContext( hpbufferdc );
  assert (hpbufferglrc);

  return TRUE;

#undef ADD_PAIR
}


void eventloop (void) {
  MSG msg;
  BOOL bRet;

  while( (bRet = GetMessage( &msg, NULL, 0, 0 )) != 0)
    {
      if (bRet == -1)
        {
          // handle the error and possibly exit
        }
      else
        {
          TranslateMessage(&msg);
          DispatchMessage(&msg);
        }
    }
}
