
/*
  guard_bits_mad : evaluate the intermediate precision of the first mad unit
*/


#include <stdio.h>
#include <assert.h>
#include <math.h>

#include "gpubench.h"
#include "arith.h"


void fill_data(float * xData, float * yData, float * zData, int components, int size)
{
    int arraySize = size * size * components;
    float tmp = -.5f;
    for(int i=0; i<arraySize; i+=4) {
		xData[i] = xData[i+1] = xData[i+2] = xData[i+3] = random_float(-5, +5, 0);
		yData[i] = yData[i+1] = yData[i+2] = yData[i+3] = random_float(-5, +5, 0);
		zData[i] = zData[i+1] = zData[i+2] = zData[i+3] = - (xData[i]+yData[i]) * xData[i];
		tmp *= 0.5;
    }
}

char const * shader_ati()
{
    return "shaders/guard_bit_mad_2/guard_bit_mad2_ati.arb";
}

char const * shader_nv()
{
    return "shaders/guard_bit_mad_2/guard_bit_mad2_nv.arb";
}

void check_results(float * results, float * xData, float * yData, float * zData, int arraySize)
{
    // For debugging purposes
    print_out("result_guard_bit_mad2.txt", results, xData, arraySize);

    int i = 1;
    double tmp;

    float * dataX = xData;
    float * dataY = yData;
    float * dataZ = zData;
    float * data = results;
    int N = arraySize;

    // check how many bits are used within the mad :
    double emin = log(abs(((((double)dataX[0] + (double)dataY[0])*(double)dataX[0] + (double)dataZ[0]) - (double)data[0])/(double)dataZ[0]))/log(2.0);
    int ind_min=0;

    while(i<N){
		tmp = log(abs(((((double)dataX[i] + (double)dataY[i])*(double)dataX[i] + (double)dataZ[i]) - (double)data[i])/(double)dataZ[i]))/log(2.0);
		//printf("i: %d %f %a %a %a \n",i, tmp,  dataZ[i], data[i], (((((double)dataX[i] * (double)dataY[i]) + (double)dataZ[i]) - (double)data[i])/(double)dataZ[i]));
		if (tmp > emin){
			emin = tmp;
			ind_min = i;
		}
		i+=4;
    }
    i=2;
    while(i<N){
		tmp = log(abs(((((double)dataX[i] + (double)dataY[i])*(double)dataX[i] + (double)dataZ[i]) - (double)data[i])/(double)dataZ[i]))/log(2.0);
		//printf("i: %d %f %a %a %a \n",i, tmp,  dataZ[i], data[i], (((((double)dataX[i] * (double)dataY[i]) + (double)dataZ[i]) - (double)data[i])/(double)dataZ[i]));
		if (tmp > emin){
			emin = tmp;
			ind_min = i;
		}
		i+=4;
    }
    i=3;
    while(i<N){
		tmp = log(abs(((((double)dataX[i] + (double)dataY[i])*(double)dataX[i] + (double)dataZ[i]) - (double)data[i])/(double)dataZ[i]))/log(2.0);
		//printf("i: %d %f %a %a %a \n",i, tmp,  dataZ[i], data[i], (((((double)dataX[i] * (double)dataY[i]) + (double)dataZ[i]) - (double)data[i])/(double)dataZ[i]));
		if (tmp > emin){
			emin = tmp;
			ind_min = i;
		}
		i+=4;
    }

    printf("The intermediate precision of the second mad within the pixel shaders mad unit seems to be : %f bits ",emin);
}

int main(int argc, char * argv[])
{
    return do_main(argc, argv, 8, pixel_shader);
}
