
/*
  mult_trunc:
  checks whether multiplication is truncated
*/


#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "gpubench.h"
#include "getopt.h"

#include "arithcommon.h"

void usage(char *name) {

    fprintf (stderr, "Usage: %s <options>\n", name);
    fprintf (stderr, "  Options\n");
    fprintf (stderr, "  -s, --size=SIZE\n");
    fprintf (stderr, "             size of the texture and framebuffer (512)\n");
    fprintf (stderr, "  -c, --components=SIZE\n");
    fprintf (stderr, "            number of components for texture and\n");
    fprintf (stderr, "            render target. (4)\n");
    fprintf (stderr, "  -i, --iterations=INT\n");
    fprintf (stderr, "            Number of iterations to perform (default: 1000)\n");
    exit(0);
}

char const * shader_ati()
{
    return "shaders/mult_error/mult_err_ati.arb";
}

char const * shader_nv()
{
    return "shaders/mult_error/mult_err_nv.arb";
}

//static float const step = (float)ldexp(1.0, -20);

void fill_data(float * xData, float * yData, float * zData, int components, int size, float step)
{
    int arraySize = size * size * components;

    float x = 1.0f + (float)ldexp(1.0, -23);
    for (int i=0;i<arraySize;i++)
    {

	xData[i] = x;
	yData[i] = 1.0f + i * step;
	zData[i] = 0.0f; // not used
    }
}

void fill_data_2(float * xData, float * yData, float * zData, int components, int size, int k, float step)
{
    int arraySize = size * size * components;

    float y = 1.0f + k * step - (float)ldexp(1.0f, -23);
    for(int i = 0; i < arraySize; i++)
    {
	if(i < 24)
	{
	    xData[i] = 1.0f + (float)ldexp(1.0, -(i+1)) + (float)ldexp(1.0, -23);
	    yData[i] = y;
	}
	else
	{
	    xData[i] = 0.0f;
	    yData[i] = 0.0f;
	}
	zData[i] = 0.0f;
    }
}

int check_results_1(float * results, float * xData, float * yData, float * zData, int arraySize, float step)
{
    int k = -1;
    for(int i = 0; i < arraySize; i++)
    {
	if(results[i] != 1.0f + i * step + (float)ldexp(1.0, -23))
	{
	    k = i;
	    break;
	}
    }

	if(k != -1) {
		printf("k=%d x %a (0x%x), x=%a, y=%a, r=%a\n", k, step, k, xData[k], yData[k], results[k]);
	    return k;
	}
	else {
		printf("No truncation???\n");
		return 0;
	}


//     printf("Computing r = 1+3*2^(-23) * yoyo\n");
//     for(int i = 0; i < arraySize; i+=2)
//     {
// 	printf("i=%d, y=%a (no carry) ->\tr=%a\n", (i/2)+1, yData[i], results[i]);
// 	printf("i=%d, y=%a (carry)    ->\tr=%a\n", (i/2)+1, yData[i+1], results[i+1]);

// //	printf("i=%d, x=%a, y=%a ->\tr=%a\t(%e)\n", i+1, xData[i], yData[i], results[i], results[i]);
//     }
}

void check_results_2(float * results, float * xData, float * yData, float * zData, int arraySize, int k, float step)
{
    printf("Computing r = (1 + 2^-i + 2^-23) * (1 + (k-1)2^-23)\n");
    for(int i = 0; i < 24; i++)
    {
	printf("i=%d, x=%a, y=%a -> r=%a\n", i+1, xData[i], yData[i], results[i]);
    }
}


static int format[5] = {0, GL_RED, GL_LUMINANCE_ALPHA, 
			GL_RGB, GL_RGBA};

int main(int argc, char** argv)
{
    //return do_main(argc, argv, 1024, pixel_shader);
    int defsize = 1024;

    int i;

    char *progname = argv[0];

    int opt;
    int opt_index;

    int size = defsize;
    int iters = 1;
    int components = 4;
    bool doVerify = true;
	float step;

	
    static struct option long_options[] = {
	{"size",         1, 0, 's'},
	{"components",   1, 0, 'c'},
	{"iters",        1, 0, 'i'},    
	{0, 0, 0, 0}
    };

    while ((opt = getopt_long(argc, argv, "s:c:i:",
			      long_options, &opt_index)) != EOF) {
	switch (opt) {
	case 's':
	    size = atoi(optarg);
	    break;
	case 'c':
	    components = atoi(optarg);
	    break;    
	case 'i':
	    iters = atoi(optarg);
	    break;
	case 'z':
	    doVerify = true;
	    break;
	default:
	    usage(progname);
	}
    }
  
    argv += optind;
    argc -= optind;
    if (argc < 0) 
	usage(progname);

    if (size < 1 || size > 2048)      
	usage(progname);
	
    int arraySize = components * size * size;
	step = 1.0f / arraySize;

		
    printf("(* size=%d components=%d iters=%d step=%g *)\n", size, components, iters, step);	
    fflush(stdout);

    gpubench_createWindow(10, 10, NULL, "Guard Bits Test");

    gpubench_createFloatPbufferComponents(size, size, GL_TRUE, components);
    gpubench_makeCurrentPbuffer();

    glDrawBuffer(GL_BACK);
    glReadBuffer(GL_BACK);
    glDisable (GL_DEPTH_TEST);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glDrawBuffer(GL_FRONT);
    glReadBuffer(GL_FRONT);
    glDisable (GL_DEPTH_TEST);
    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    shader_t multFP;
    multFP = gpubench_loadShader(gpubench_isNV() ? shader_nv() : shader_ati());


    if (multFP == GPUBENCH_BOGUS_PROGRAM) {
	fprintf(stderr, "Error, couldn't load program.\n");
	gpubench_destroyPbuffer();
	exit(1);
    }


    float* xData = (float*)malloc(sizeof(float) * arraySize);
    float* yData = (float*)malloc(sizeof(float) * arraySize);
    float* zData = (float*)malloc(sizeof(float) * arraySize);

    GLuint texId[3];

    fill_data(xData, yData, zData, components, size, step);

    texId[0] = gpubench_makeTex(0, size, size, components, xData);
    texId[1] = gpubench_makeTex(1, size, size, components, yData);
    texId[2] = gpubench_makeTex(2, size, size, components, zData);


    gpubench_bindShader(multFP);
	
    int draw_to_front = 1;

    float scaleX = (gpubench_isNV()) ? 1.0f : (float)size;
    float scaleY = (gpubench_isNV()) ? 1.0f : (float)size;

    // force load of textures onto card by rendering geometry
    glBegin(GL_TRIANGLES);    
    glTexCoord2f(0.0f, size*2.0f / scaleY);      
    glVertex2f(-1.0f, 3.0f); 
    glTexCoord2f(0.0f, 0.0f);      
    glVertex2f(-1.0f, -1.0f);   
    glTexCoord2f(2.0f*size / scaleX, 0.0f);      
    glVertex2f(3.0f, -1.0f);
    glEnd();

    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    gpubench_resetTimer();

    for (i=0;i<iters; i++) {

	glBegin(GL_TRIANGLES);    
	glTexCoord2f(0.0f, size*2.0f / scaleY);      
	glVertex2f(-1.0f, 3.0f); 
	glTexCoord2f(0.0f, 0.0f);      
	glVertex2f(-1.0f, -1.0f);   
	glTexCoord2f(2.0f*size / scaleX, 0.0f);      
	glVertex2f(3.0f, -1.0f);
	glEnd();

	draw_to_front = gpubench_bindAndSwap(draw_to_front);
    }

    glFinish();
	
    float ms = gpubench_getTimer();

    CHECK_GL();

    draw_to_front = gpubench_bindAndSwap(draw_to_front);
  
    int numErrors = 0;
    float* results = (float*)malloc(sizeof(float) * arraySize);

    glReadPixels(0, 0, size, size, format[components], GL_FLOAT, results);

    int k=check_results_1(results, xData, yData, zData, arraySize, step);


    // second pass
    fill_data_2(xData, yData, zData, components, size, k, step);

    gpubench_sendToGPU(texId[0], size, size, components, xData);
    gpubench_sendToGPU(texId[1], size, size, components, yData);
    gpubench_sendToGPU(texId[2], size, size, components, zData);


    gpubench_render(size, size, 0, 1, iters, GPUBENCH_ACCESS_SEQ);

    glReadPixels(0, 0, size, size, format[components], GL_FLOAT, results);

    check_results_2(results, xData, yData, zData, arraySize, k, step);

    free(results);



    printf("%4.4f  msec per iter  (%4.4f msec total)\n", ms / 1000.0f, ms);
    printf("%4.4f  MFLOPS\n", (2.0f * iters * components * size * size) / ms / 1000.0f);

    free(xData);
    free(yData);
    free(zData);

    glDeleteTextures(3, texId);
    gpubench_destroyPbuffer();

    return 0;
}
